# /u/beebe/adobe/fpp.awk, Wed Dec 26 18:39:08 1990
# Edit by Nelson H. F. Beebe <beebe@ibm>
# ======================================================================
# This is a simple preprocessor for Fortran and SFTRAN3, modelled
# on the C preprocessor, but tailored for Fortran use, and for
# reversibility.
#
# Complete documentation can be found in the manual page, fpp(1).
#
# Usage:
#	nawk -f fpp.awk
#               [+debug]
#               [+version]
#		[+Dname1] [+Dname2=value]
#		[+Uname3]
#		[name1=value1] ... [namek=valuek]
#		[--]
#		[infile1] ... [infilek] >outfile
#
# The GNU implementation of nawk, called gawk, may be used as
# well, provided it is invoked as with the compatibility option,
# -c.  Normally, this would be hidden inside a script so that it
# could be invoked by
#
#	fpp macro1=value1 ... macrok=valuek <infile >outfile
#
# [26-Dec-1990]
# ======================================================================
BEGIN {
	initialize();
}

# ======================================================================

/^[Cc*]-fpp-/ {
# printf("-fpp- line: selected[%d] = %d [%s]\n",level,selected[level],$0);
		next;		# discard old preprocessor id lines
	}

# preprocessor line
/^[Cc*][ \t]*#[ \t]*[A-Za-z]/ {

# printf("FPP line: selected[%d] = %d [%s]\n",level,selected[level],$0);

	    preprocessor_line();
	    next;		# so we ignore remaining patterns
	}


/^[Cc*][^#]*#[ \t]*\(/ {
# printf("EXPMAC line: selected[%d] = %d [%s]\n",level,selected[level],$0);
	    expand_macros();
	    next;		# so we ignore remaining patterns
	}



	{
# printf("NORMAL line: selected[%d] = %d [%s]\n",level,selected[level],$0);
	    normal_line();	# all non-preprocessor lines
	}

END	{			# Check for conditional nesting errors
		if (level > 0)
		    error("?Open conditionals at input lines:",1);
		else if (level < 0)
		    error("?Too many #endifs at input line:",1);
		exit exit_code;
	}

# ======================================================================

function abort(message,dump_lineno)
{
    error(message,dump_lineno);
    err_printf(\
	sprintf("Processing terminated at input line %d [%s]\n",FNR,$0));
    exit 1;
}


function do_def(name,value,perm_level, trimmed_value)
{
    if (name !~ /^[A-Za-z_][A-Za-z_0-9]*$/)
	error("Symbol name illegal: does not match ^[A-Za-z_][A-Za-z_0-9]*$");
    trimmed_value = value
    sub(/^[ \t]+/,"",trimmed_value);	# remove leading whitespace
    sub(/[ \t]+$/,"",trimmed_value);	# remove trailing whitespace
    if (trimmed_value == "")
	trimmed_value = "1";
    if ((substr(name,1,2) == "__") && (perm_level < 1))
    {
	error(\
      sprintf("Permanent symbol %s at input line %d cannot be redefined\n",\
	    name,FNR),0);
    }
    else
    {
	if (substr(name,1,2) != "__") # do not show permanent macros
	    printf("C-fpp- Macro: %s=%s\n",name,trimmed_value);
	if (debug)		# debugging shows them all
	    err_printf(sprintf("DEBUG: Macro: %s=%s\n",name,trimmed_value));
	definition[name] = trimmed_value;
    }
}


function do_identify()
{
    do_def("__FILE__",FILENAME,1); # changes dynamically
    do_def("__LINE__",FNR,1);	# changes dynamically
}


function do_undef(name)
{
# print "UNDEFINING",name;
    if (name !~ /^[A-Za-z_][A-Za-z_0-9]*$/)
	error("Symbol name illegal: does not match ^[A-Za-z_][A-Za-z_0-9]*$");
    if (substr(name,1,2) == "__")
	error(sprintf("Permanent symbol %s cannot be undefined\n",name),0);
    else
    {
	if (substr(name,1,2) != "__") # do not show permanent macros
	    printf("C-fpp- Macro: %s=--UNDEFINED--\n",name);
	if (debug)		# debugging shows them all
	    err_printf(sprintf("DEBUG: Macro: %s=--UNDEFINED--\n",name));
	delete definition[name];
    }
}


function error(message,dump_lineno, k)
{
    err_printf(identification() "?? " message);
    if (dump_lineno && (level > 0))
    {
	for (k = 1; k <= level; ++k)
	    err_printf(sprintf(" %d",lineno[k]));
    }
    else if (dump_lineno && (level < 0))
    {
	for (k = 1; k > level; --k)
	    err_printf(sprintf(" %d",lineno[k]));
    }
    err_printf("\n");
    err_printf(sprintf("Input line = [%s]\n",$0));
    exit_code = 1;
}


function err_printf(s)	# print (possibly partial line) string s on stderr
{
    printf("%s",s) | "cat 1>&2";
}


function eval(s, e,ne,temp)	# evaluate an expression, s
{
    if (debug)
	err_printf(sprintf("DEBUG: Evaluating [%s]...\n",s));
    temp = s;
    gsub(/\+/,   " + ",temp);	# ensure space separates all tokens so
    gsub(/-/,   " - ",temp);	# split() can parse the expression
    gsub(/\//,  " / ",temp);
    gsub(/\*/,  " * ",temp);
    gsub(/\*[ \t]*\*/," ** ",temp);
    gsub(/%/,   " % ",temp);
    gsub(/!/,   " ! ",temp);
    gsub(/![ \t]*=/,  " != ",temp);
    gsub(/\=\=/,  " == ",temp);
    gsub(/</,  " < ",temp);
    gsub(/<[ ]*=/,  " <= ",temp);
    gsub(/>/,  " > ",temp);
    gsub(/>[ ]*=/,  " >= ",temp);
    gsub(/&&/," \\& ",temp);	# reduce && to &
    gsub(/&/,  " \\& ",temp);
    gsub(/<[ \t]*</,"<<",temp);
    gsub(/>[ \t]*>/,">>",temp);
    gsub(/\|\|/," | ",temp);	# reduce || to |
    gsub(/\|/,  " | ",temp);
    gsub(/\(/,  " ( ",temp);
    gsub(/\)/,  " ) ",temp);
    gsub(/\/[ \t]*\//,  " // ",temp);

    gsub(/\.[Nn][Oo][Tt]\./," ! ",temp); # change Fortran .logop. to synonyms
    gsub(/\.[Aa][Nn][Dd]\./," \& ",temp);
    gsub(/\.[Oo][Rr]\./," | ",temp);
    gsub(/\.[Ee][Qq][Vv]\./," .eqv. ",temp);
    gsub(/\.[Nn][Ee][Qq][Vv]\./," .neqv. ",temp);
    gsub(/\.[Ee][Qq]\./," == ",temp);
    gsub(/\.[Nn][Ee]\./," != ",temp);
    gsub(/\.[Ll][Tt]\./," < ",temp);
    gsub(/\.[Ll][Ee]\./," <= ",temp);
    gsub(/\.[Gg][Tt]\./," > ",temp);
    gsub(/\.[Gg][Ee]\./," >= ",temp);

    gsub(/[ \t]+/," ",temp);	# normalize spacing for cleaner debug output

    ne = split(temp,ee);	# ee and ke are globals
    ke = 1;			# point to first element
    e = expr();
    if (debug)
	err_printf(sprintf("\tDEBUG: Expression [%s] -> [%s]\n",temp,e));
    if (ke < ne)
	error("Garbage follows expression",0);
    return e;
}


function expand_macros()
{
    print $0;			# preserve this comment line
    if (selected[level])
    {				# output its uncommented expansion
	do_identify();		# in case of error messages
	line = substr($0,2);
	while (match_expr(line) > 0)
	{
	line = substr(line,1,RSTART-1) \
	    eval(substr(line,RSTART + 1,RLENGTH - 1)) \
	    substr(line,RSTART + RLENGTH);
	}
	print line;
	getline line;		# discard next input line
	print "C-fpp-" line;	# but preserve it as a comment
    }
}


function expr( e,op)	# term | term [+ - | & == != <= >= < >]
{
    e = term();
    while (1)
    {
	op = ee[ke++];
	if (op == "+")
	    e += term();
	else if (op == "-")
	    e -= term();
	else if (op == "|")
	    e = e || term();
	else if (op == "&")
	    e = e && term();
	else if (op == ".neqv.")
	    e = xor(e,term());
	else if (op == ".eqv.")
	    e = !xor(e,term());
	else if (op == "==")
	    e = (e == term());
	else if (op == "!=")
	    e = (e != term());
	else if (op == "<=")
	    e = (e <= term());
	else if (op == ">=")
	    e = (e >= term());
	else if (op == "<")
	    e = (e < term());
	else if (op == ">")
	    e = (e > term());
	else
	{
	    ke--;		# just have bare term; backup one entry
	    break;
	}
    }
    return e;
}


function factor( e,k,keorg)	# number | ( expr ) | [+-!] factor |
{				# defined name | defined ( name )
    keorg = ke;			# for debug output
    if (ee[ke] == "!")		# unary not
    {
	ke++;
	e = !expr();
    }
    else if (ee[ke] == "-")	# unary minus
    {
	ke++;
	e = -expr();
    }
    else if (ee[ke] == "+")	# unary plus
    {
	ke++;
	e = expr();
    }
    else if (ee[ke] ~ /[Dd][Ee][Ff][Ii][Nn][Ee][Dd]/)
    {
	ke++;
	if (ee[ke] == "(")	# defined ( name )
	{
	    ke++;
	    if (debug)
		err_printf(sprintf("\t\t\tDEBUG: defined(%s) -> [%s]\n",\
		    ee[ke],definition[ee[ke]]));
	    e = (definition[ee[ke++]] != "");
	    if (ee[ke++] != ")")
		error("Missing right parenthesis in defined(name)",0);
	}
	else			# defined name
	{
	    if (debug)
		err_printf(sprintf("\t\t\tDEBUG: defined(%s) -> [%s]\n",\
		    ee[ke],definition[ee[ke]]));
	    e = (definition[ee[ke++]] != "");
	}
    }
    else if (ee[ke] == "(")
    {
	ke++;
	e = expr();
	if (ee[ke++] != ")")
	    error("Missing right parenthesis in expression",0);
    }
    else
    {
	if (ee[ke] ~ /^[+-]?([0-9]+[.]?[0-9]*|[.][0-9]+)([DdEe][+-]?[0-9]+)?$/)
	    e = ee[ke++];	# number
	else if (ee[ke] ~ /\.[Tt][Rr][Uu][Ee]\./) # .true.
	{
	    ke++;
	    e = 1;
	}
	else if (ee[ke] ~ /\.[Ff][Aa][Ll][Ss][Ee]\./) # .false.
	{
	    ke++;
	    e = 0;
	}
	else			# must be name; get its definition
	    e = definition[ee[ke++]];
    }
    if (debug)
    {
	err_printf("\t\tDEBUG: Factor:");
	for (k = keorg; k < ke; ++k)
	    err_printf(sprintf(" ee[%d] = [%s]",k,ee[k]));
	err_printf(sprintf(" -> [%s]\n",e));
    }
    return e;
}


function identification( temp)
{
    do_identify();
    temp = "__FILE__=" FILENAME "\t__LINE__=" FNR "\t";
    return temp;
}


function ifix(x)		# gawk bug workaround
{
    return (x >= 0) ? int(x) : -int(-x);
}


function initialize()
{
    VERSION = "fpp -- awk version 1.0 [26-Dec-1990]\n";
				# update when fpp changes!

    # regexp patterns for fpp directives
    REGEXP_DEFINE = "#[ \t]*[Dd][Ee][Ff][Ii][Nn][Ee][ \t]+";
    REGEXP_DIRECTIVE = "^[Cc*][ \t]*#[ \t]*[A-Za-z]+[ \t]*";
    REGEXP_ELSE = "#[ \t]*[Ee][Ll][Ss][Ee]([^A-Za-z]|$)";
    REGEXP_ELSEIF = "(#[ \t]*[Ee][Ll][Ii][Ff][^A-Za-z])|(#[ \t]*[Ee][Ll][Ss][Ee][Ii][Ff][^A-Za-z])";
    REGEXP_ENDIF = "#[ \t]*[Ee][Nn][Dd][Ii][Ff]([^A-Za-z]|$)";
    REGEXP_ERROR = "#[ \t]*[Ee][Rr][Rr][Oo][Rr][ \t]+";
    REGEXP_IF = "#[ \t]*[Ii][Ff][^A-Za-z]";
    REGEXP_IFDEF = "#[ \t]*[Ii][Ff][Dd][Ee][Ff][^A-Za-z]";
    REGEXP_IFNDEF = "#[ \t]*[Ii][Ff][Nn][Dd][Ee][Ff][^A-Za-z]";
    REGEXP_MESSAGE = "#[ \t]*[Mm][Ee][Ss][Ss][Aa][Gg][Ee][ \t]+";
    REGEXP_NAME = "[A-Za-z_][A-Za-z_0-9]*";
    REGEXP_UNDEFINE = "(#[ \t]*[Uu][Nn][Dd][Ee][Ff][ \t]+)|(#[ \t]*[Uu][Nn][Dd][Ee][Ff][Ii][Nn][Ee][ \t]+)";

    branch_done[0] = 1;
    debug = 0;			# turn on with -debug option
    exit_code = 0;
    IGNORECASE = 1;		# gawk only; makes regexps ignore letter case
    level = 0;			# preprocessor statement nesting level
    selected[0] = 1;		# normally, all lines are selected
    "echo $USER" | getline user;
    "echo $HOSTNAME" | getline host;
    "pwd"  | getline cwd;
    "date" | getline date;
    print "C-fpp- =================================================================";
    printf("C-fpp- %s",VERSION);
    printf("C-fpp- Date: %s\n",date);
    printf("C-fpp- Directory: %s\n",cwd);
    printf("C-fpp- User: %s%s%s\n",user,(host == "") ? "" : "@",host);
    do_def("__DATE__",substr(date,5,3) substr(date,8,3) substr(date,24,5),1);
    do_def("__TIME__",substr(date,12,8),1);
    do_def("__TIMEZONE__",substr(date,21,3),1);
    for (k = 1; k < ARGC; ++k)
    {
	if (ARGV[k] == "--") # end of non-file arguments
	{
	    ARGV[k] = "";	# clear so switch is not treated as a filename
	    break;		# stop scanning ARGV[]
	}
	else if (ARGV[k] == "+debug")
	{
	    debug = 1;
	    ARGV[k] = "";	# clear so switch is not treated as a filename
	}
	else if (ARGV[k] == "+version")
	{
	    err_printf(VERSION);
	    ARGV[k] = "";	# clear so switch is not treated as a filename
	}
	else if (substr(ARGV[k],1,2) == "+D")
	{			# allow cc-style +Dname and +Dname=var
	    ARGV[k] = substr(ARGV[k],3) \
		((index(ARGV[k],"=") > 0) ? "" : "=1");
	}
	else if (substr(ARGV[k],1,2) == "+U")
	{			# allow cc-style +Uname
	    undef = 1;
	    sub(/\=.*$/,"",ARGV[k]); # discard any explicit =value field
	    do_undef(substr(ARGV[k],3));
	    ARGV[k] = "";
	}
	m = index(ARGV[k],"=");
	if (m > 0)	# save macro values indexed by macro name
	{
	    do_def(substr(ARGV[k],1,m-1),substr(ARGV[k],m+1),0);
	    ARGV[k] = "";	# clear so +Dname or +Uname are not treated as
				# filenames.
	}
    }
    print "C-fpp- =================================================================";
}

function lshift(a,n)	# left shift (fails for negative a)
{
    for (; n > 0; --n)
	a = int(a) * 2;
    return (a);
}


function match_expr(e, k,plevel) # match balanced expression #(...)
{				# setting RSTART and RLENGTH
#    if (debug)
#	err_printf(sprintf("DEBUG: match_expr(%s) -> ",e));
    if (match(e,/#[ \t]*\(/) == 0)
	return RSTART;
    plevel = 1;
    k = RSTART + RLENGTH;
    while ((k <= length(e)) && (plevel != 0))
    {
	if (substr(e,k,1) == "(")
	    plevel++;
	else if (substr(e,k,1) == ")")
	    plevel--;
	k++;
    }
    if (plevel == 0)
	RLENGTH = k - RSTART;
    else
    {
	error(\
	  sprintf("Unbalanced parenthesis in #(...) expression at line %d\n",\
	    NFR),0);
	RSTART = 0;
	RLENGTH = 0;
    }
#   if (debug)
#	err_printf(sprintf("%d %d\n",RSTART,RLENGTH));
    return RSTART;
}


function normal_line()
{
    if (level > 0)
    {
	if (selected[level])
	    print (substr($0,1,3) == "C##") ? substr($0,4) : $0;
	else
	    print (substr($0,1,3) == "C##") ? $0 : "C##" $0;
    }
    else
	print $0;
}


function preprocessor_line()
{
    match($0,REGEXP_DIRECTIVE);
    if ($0 ~ REGEXP_IF)
    {
	level++;
	lineno[level] = FNR;	# global value for error messages
	selected[level] = selected[level-1] && \
	    eval(substr($0,RSTART + RLENGTH));
	branch_done[level] = selected[level];
    }
    else if ($0 ~ REGEXP_IFDEF)
    {
	level++;
	lineno[level] = FNR;	# global value for error messages
	selected[level] = selected[level-1] && \
	     (definition[substr($0,RSTART + RLENGTH)] != "");
	branch_done[level] = selected[level];
    }
    else if ($0 ~ REGEXP_IFNDEF)
    {
	level++;
	lineno[level] = FNR;	# global value for error messages
	selected[level] = selected[level-1] && \
	    (definition[substr($0,RSTART + RLENGTH)] == "");
	branch_done[level] = selected[level];
    }
    else if ($0 ~ REGEXP_ELSEIF)
    {
	if (level == 0)
	    error("No preceding #if",0);
	lineno[level] = FNR;	# global value for error messages
	selected[level] = selected[level-1] && \
	    eval(substr($0,RSTART + RLENGTH));
	if (!branch_done[level])
	    branch_done[level] = selected[level];
    }
    else if ($0 ~ REGEXP_ELSE)
    {
	if (level == 0)
	    error("No preceding #if",0);
	selected[level] = selected[level-1] && !branch_done[level];
    }
    else if ($0 ~ REGEXP_ENDIF)
    {
	if (level == 0)
	    error("No preceding #if",0);
	lineno[level] = FNR;
	level--;
    }
    else if (selected[level])
    {

#   print "DEBUG: selected:",$0;
#   print "DEBUG: message pattern = [" REGEXP_MESSAGE "]";

	if (match($0,REGEXP_DEFINE) > 0)
	{
	    line = substr($0,RSTART + RLENGTH);
	    match(line,REGEXP_NAME);
	    do_def(substr(line,RSTART,RLENGTH),\
		substr(line,RSTART + RLENGTH),0);
	}
	else if (match($0,REGEXP_UNDEFINE) > 0)
	{
	    line = substr($0,RLENGTH + 1);
	    match(line,REGEXP_NAME);
	    do_undef(substr(line,RSTART,RLENGTH));
	}
	else if (match($0,REGEXP_MESSAGE) > 0)
	{
# print "MESSAGE: ",$0;

	    errmsg = identification() substr($0,RSTART + RLENGTH) "\n";
	    err_printf(errmsg);
#    	    err_printf(identification() substr($0,RSTART + RLENGTH) "\n");
	}
	else if (match($0,REGEXP_ERROR) > 0)
	{
	    print identification() substr($0,RSTART + RLENGTH);
	    error(substr($0,RSTART + RLENGTH) "\n",0);
	}
	else
	{
	    print identification() "?? Unrecognized:",$0;
	    error("Unrecognized:" $0,0);
	}
    }
    print $0;
}


function rshift(a,n)	# right shift (fails for negative a)
{
    for (; n > 0; --n)
    {
	a = int(a);
	a /= 2;
    }
    return (a);
}


function term ( e,f,g)	# factor | factor [* / % ** //] factor
{
    e = factor();
    while (1)
    {
	f = ee[ke++];
	if (f == "*")
	    e *= factor();
	else if (f == "/")
	    e /= factor();
	else if (f == "%")	# use Fortran's definition instead of
	{			# awk's system-dependent one
	    f = factor();
	    g = ifix(e/f);	# gawk bug workaround; cannot use int()
	    e = ifix(e - g*f);
	}
	else if (f == "**")
	    e ^= factor();	# ^ is awk's exponeniation operator
	else if (f == "//")	# Fortran concatenation
	    e = e factor();
	else if (f == "<<")
	    e = lshift(e,factor());
	else if (f == ">>")
	    e = rshift(e,factor());
	else
	{
	    ke--;		# no following operator, so backup
	    break;
	}
    }
    return e;
}


function xor(a,b)
{
    return ((a && !b) || (!a && b)) ? 1 : 0;
}
